// ==================================================================== 
// 
//  Geodetic.cc; a part of geo, a geodetic computer for amateur radio
//               operators.
// 
//  Copyright (C) 1999David Freese                                  
//                                                                    
//  This program is free software; you can redistribute it and/or     
//  modify itunder the terms of the GNU General Public License as    
//  published by the Free Software Foundation; either version 2 of    
//  the License, or (at your option) any later version.              
//                                                                    
//  This program is distributed in the hope that it will be useful,   
//  but WITHOUT ANY WARRANTY; without even the implied warranty of    
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     
//  GNU General Public License for more details.                     
//                                                                    
//  You should have received a copy of the GNU General Public License 
//  along with this program; if not, write to the Free Software       
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
//                                                                    
//  Author: Dave Freese                                             
//  Internet: dfreese@intrepid.net                                   
//                                                                    
//====================================================================

// Class library for computing the bearing & distance between two points
// on the surface of the earth.
// Default Earth model is a sphere.
// Other earth models :
//   oblate spheroid, Andoyer Lambert model,
//   oblate spheroid, Sadonis(5th order) model.

#include <math.h>
#include "geodetic.h"

#define EPSILON        1.0e-9
#define PI             M_PI
#define TPI            (2.0 * PI)
#define PI2            (PI / 2.0)
#define CONV           (PI / 180.0)
#define MINUTE         (CONV / 60.0)
//#define MAJORAXIS      3437.7468          // nautical miles 
//#define MINORAXIS      3426.2208          // nautical miles 
//#define MAJORAXIS      3906.5305          // statute miles 
//#define MINORAXIS      3893.4327          // statute miles 
#define MAJORAXIS      6286.9513          // kilometers 
#define MINORAXIS      6265.8726          // kilometers


// constants used for Andoyer Lambert Earth Model 

#define  Ecc 0.99664721741   //  eccentricity = Minor / Major
#define  k1  0.00336402314   // (1 - Ecc)*(2 - Ecc) 
#define  k2  0.00168203047   // (A/B) + (A/B)^2 + (A/B)^3 
#define  k3  0.00167639100   // (A/B) - (A/B)^2 + (A/B)^3 
#define  k4  1.1241147e-5    //  A*A 
#define  k5  1.1316650e-5    //  k1*k1 

double SodanoAngle();
double SodanoDistance();

geodetic :: geodetic( int iModel) {
  Lat1 = PI2;
  Lat2 = PI2;
  coLat1 = 0;
  coLat2 = 0;
  Lon1 = 0;
  Lon2 = 0;
  bearing = 0;
  distance = 0;
  model = iModel;
}

geodetic :: geodetic( double lat1, double lon1, double lat2, double lon2,
                      int iModel ) {
  Lat1 = lat1*CONV; Lat2 = lat2*CONV;
  coLat1 = PI2 - Lat1;
  Lon1 = lon1*CONV;
  coLat2 = PI2 - Lat2;
  Lon2 = lon2*CONV;
  bearing = 0.0;
  distance = 0.0;
  model = iModel;
  compute();
}

void geodetic :: P1( double Lat, double Lon ) {
  Lat1 = Lat*CONV;
  coLat1 = PI2 - Lat1;
  Lon1 = -Lon*CONV;  // models are base on +n, +w, but use +n, +e for data input
  compute();
}

void geodetic :: P2( double Lat, double Lon ) {
  Lat2 = Lat*CONV;
  coLat2 = PI2 - Lat2;
  Lon2 = -Lon*CONV;
  compute();
}

void geodetic :: Pairs( double lat1, double lon1, double lat2, double lon2 ) {
  Lat1 = lat1*CONV;
  Lat2 = lat2*CONV;
  coLat1 = PI2 - Lat1;
  Lon1 = -lon1*CONV;
  coLat2 = PI2 - Lat2;
  Lon2 = -lon2*CONV;
  compute();
}

int geodetic :: GetModel()
{
  return( model );
}

void geodetic :: SetModel( int m )
{
  model = m;
}

double geodetic :: Bearing() {
  return( bearing );
}

double geodetic :: Distance() {
  return( distance );
}

void geodetic :: compute()
{
  switch ( model ) {
    case SPHERE : compute_Sphere();
                  break;
    case ANDOYER : compute_Andoyer();
                   break;
    case SODANO : compute_Sodano();
                   break;
    default : compute_Sphere();
  }
}

//  Spherical Model computation

void geodetic :: compute_Sphere() {
  double costest;
  double sinof;

  diff = (Lon1 - Lon2);

  if( diff > PI )
    diff -= TPI;
  else if( diff < -PI )
    diff += TPI;

  v = acos( cos(coLat1)*cos(coLat2) + sin(coLat1)*sin(coLat2)*cos(diff) );

  if( fabs(diff) <= EPSILON ) {
    if( coLat1 >= coLat2 )
      bearing = 0.0;
    else
      bearing = 180.0;
  } else { 
      sinof = sin(coLat2) * sin(diff) / sin(v);
      if( sinof > 1 ) sinof = 1;
      if( sinof < -1) sinof = -1;
      
      bearing = asin( sinof ) / CONV;

      costest = cos(coLat2) - cos(coLat1)*cos(v);

      if( costest < 0  && bearing >=0 )
        bearing = 180 - bearing;
      if( costest < 0 && bearing < 0 )
        bearing = 180 - bearing;
      if( costest >= 0 && bearing < 0 )
        bearing = 360 + bearing;
  }
  distance = v * MAJORAXIS;
}

double fsin( double a, double b, double c, double d, double e ) 
{ 
  double f; 
  f = sin(e/2); 
  f = sin(a) + 2*b*cos(c)*sin(d)*f*f; 
  if (f == 0) 
    f = 1.0e-20; 
  return (f); 
} 
 
double fsin2( double a, double b, double c, double d, double e ) 
{ 
  double f; 
  f = fsin(a,b,c,d,e); 
  return( f*f ); 
} 
 
double arcsin( double a ) 
{ 
  return( atan( a / ( 1 - a * a ) ) ); 
} 
 
double arccos( double a ) 
{ 
  return( atan((1-a*a)/a) ); 
} 
 
// Computation based on Andoyer-Lambert earth model approximation
// this model is untested

void geodetic :: compute_Andoyer()
{
double 
  DLon,
  LonA,
  LonB, 
  RLatA, 
  RLatB, 
  angle,
  Azmth,
  p1, p2, p3,p4,p5,p6,p7, 
  p8,p10,p11,p12,p13,p14,p9,p62; 

  LonA = -Lon1;
  LonB = -Lon2;

  DLon = LonB - LonA; 

  if (DLon > PI ) 
    DLon = DLon - TPI; 

  if ( DLon < -PI ) 
    DLon = DLon + TPI; 

  if ( DLon == 0 ) 
    DLon = EPSILON; 

  RLatA = atan(Ecc*tan(Lat1)); 

  RLatB = atan(Ecc*tan(Lat2)); 

  p1 = sin(RLatA)*sin(RLatB); 

  p2 = cos(RLatA)*cos(RLatB); 

  p3 = Lat2 - Lat1 + 2*sin(Lat2-Lat1)*(p1*k2 - p2*k3); 

  p4 = sin(DLon)*cos(RLatB); 

  p4 = sqrt(p4*p4 + fsin2(p3,1,RLatB,RLatA,DLon)); 

  p5 = p1 + p2*cos(DLon); 

  if ( fabs(p4) < fabs(p5) ) 
    p6 = arcsin(p4); 
  else 
    p6 = arccos(p5); 

  p62 = p6*p6; 

  p7 = p2*sin(DLon)/sin(p6); 

  p8 = 1.0 - p7*p7; 

  p9 = sin(p6)*cos(p6); 

  p10 = p8*k5*(p9/4 - 1.25*p6 + p62/tan(p6)); 

  p10 = DLon + p7*(k1*p6 - k4*p1*(sin(p6)/2 + p62/sin(p6)) + p10); 
 
  angle = atan(cos(RLatB)*sin(p10)/fsin(p3,1,RLatB,RLatA,p10)); 
 
  if ( DLon < 0 ) 
    Azmth = angle + PI; 
  else 
    Azmth = angle; 
 
  if ( Azmth < 0 ) 
    Azmth = Azmth + PI; 

  p11 = (1+k1)*p6 + p1*(k1*sin(p6) - k4*p62/(2*sin(p6))); 

  p12 = p8*(k4*p62*cos(p6)/(2*sin(p6)) - k1*(p6 + p9)/2); 

  p13 = p1*(p8*(p62/sin(p6) + p9*cos(p6)) - p1*p9); 

  p14 = p8*p8*((p6 + p9)/8 - p62/tan(p6) - p9*cos(p6)*cos(p6)/4); 

  distance = (p11 + p12 + k4*(p13 + p14)/2);

  if ( distance < 0 )
    distance += PI;

  distance *= MAJORAXIS; 

  bearing = Azmth / CONV;
}

// Computation of great-circle distance based on 
// Sodano's 5th Order Earth Model 
// this model is untested
 
void geodetic :: compute_Sodano()
{
  distance = SodanoDistance();
  bearing = SodanoAngle() / CONV;
}

double geodetic :: SodanoDistance()
{ 
double
  LonA, LonB, LatA, LatB, 
  Pv3, Pv4, Pv5, Pv6, Pv7, Pv8, Pv9, 
  Pv10, Pv11, Pv12, Pv13, Pv14, Pv15, Pv16, Pv17, Pv18, Pv19, 
  Pv20, Pv21, Pv22, Pv23, 
  F, M, L; 

  LonA = -Lon1; LonB = -Lon2;
  LatA = Lat1; LatB = Lat2;
 
  L = 1.0 - Ecc; 
  M = 1.0 + L + L*L; 
  F = L*L/2.0; 
 
  Pv3 = atan( sin( LatB)/cos(LatB)/Ecc ); 
  Pv4 = sin(Pv3); 
  Pv5 = cos(Pv3); 
  Pv6 = LonB-LonA; 
  Pv3 = atan( sin(LatA)/cos(LatA)/Ecc ); 
  Pv7 = sin(Pv3); 
  Pv8 = cos(Pv3); 
  Pv9 = fabs(Pv6); 
  if (Pv9 - PI >= 0 ) 
    Pv10 = 2*PI - Pv9; 
  else 
    Pv10 = Pv9; 
 
  Pv11 = sin(Pv10); 
  Pv12 = cos(Pv10); 
  Pv13 = Pv4*Pv7 + Pv5*Pv8*Pv12; 
  Pv14 = sqrt( (Pv11*Pv8)*(Pv11*Pv8) + 
  	 (Pv7*Pv5 - Pv4*Pv8*Pv12)*(Pv7*Pv5 - Pv4*Pv8*Pv12) ); 
  Pv15 = 1.0 - (Pv5*Pv8*Pv11/Pv14)*(Pv5*Pv8*Pv11/Pv14); 
  Pv16 = atan(Pv14/sqrt(1.0 - Pv14*Pv14)); 
  Pv17 = Pv16*Pv16; 
  Pv18 = M*Pv16 + Pv4*Pv7*( (M-1)*Pv14 - F*Pv17/Pv14 ); 
  Pv19 = Pv15*( (1-M)/2*(Pv16+Pv14*Pv13)+F*Pv17*Pv13/Pv14); 
  Pv20 = -(Pv4*Pv7)*(Pv4*Pv7)*F*Pv13*Pv14; 
  Pv21 = ( L*L/16*(Pv13*Pv14+Pv16) - 
  	 F*Pv17*Pv13/Pv14 - F/4*Pv14*Pv13*Pv13*Pv13 ) * Pv15*Pv15; 
  Pv22 = Pv4*Pv7*Pv15*F*(Pv17/Pv14+Pv14*Pv13*Pv13); 
  Pv23 = (Pv18+Pv19+Pv20+Pv21+Pv22); 

  return (Pv23 * MAJORAXIS);

} 
 

double geodetic :: SodanoAngle()
{ 
double 
  LatA, LonA, LatB, LonB,
  Pv3, Pv4, Pv5, Pv6, Pv7, Pv8, Pv9, 
  Pv10, Pv11, Pv12, Pv13, Pv14, Pv15, Pv16, Pv17, Pv18, Pv19, 
  Pv20, Pv21, Pv22, Pv23, Pv24, Pv25, Pv26, Pv27, Pv28, pa, 
  F, L, M; 

  LatA = Lat1; LonA = -Lon1;
  LatB = Lat2; LonB = -Lon2;
 
  L = 1.0 - Ecc; 
  M = 1.0 + L + L*L; 
  F = L*L/2.0; 

  Pv3 = (MINORAXIS - MAJORAXIS) / (MINORAXIS + MAJORAXIS); 
  Pv4 = fabs( LonB - LonA ); 
  if( Pv4 == 0.0 && ( LatB - LatA == 0.0 ) ) 
    return 0.0; 
  if( Pv4 == 0.0 ) 
    Pv4 = 1.E-9; 
  if( LonB - LonA >= 0.0 ) { 
    if( Pv4 - PI <= 0.0 ) { 
      Pv5 = 1.0; 
      Pv6 = LatB; 
      Pv7 = LatA; 
    } else { 
      Pv5 = 0.0; 
      Pv6 = LatA; 
      Pv7 = LatB; 
    } 
  } else { 
    if( Pv4 - PI > 0.0 ) { 
      Pv5 = 1.0; 
      Pv6 = LatB; 
      Pv7 = LatA; 
    } else{ 
      Pv5 = 0.0; 
      Pv6 = LatA; 
      Pv7 = LatB; 
    } 
  } 
  if( Pv4 - PI >= 0.0 ) 
    Pv4 = 2.0*PI-Pv4; 
 
  Pv12 = Pv4; 
  Pv8 = atan(sin(Pv6)/cos(Pv6)*(1-L)); 
  Pv9 = atan(sin(Pv7)/cos(Pv7)*(1-L)); 
  Pv10 = sin(Pv8)*sin(Pv9); 
  Pv11 = cos(Pv8)*cos(Pv9); 
  Pv13 = fabs(Pv10+Pv11*cos(Pv12)); 
  Pv14 = Pv7-Pv6; 
  Pv15 = sin( Pv14 + 
         2.0*sin(Pv14) * ( Pv10 * (Pv3 + Pv3*Pv3 + Pv3*Pv3*Pv3 ) 
															- Pv11 * (Pv3- Pv3*Pv3 + Pv3*Pv3*Pv3 ) ) ); 
  Pv16 = sin(Pv12)*cos(Pv7); 
  pa = Pv15 + 2.0 * cos(Pv9) * sin(Pv8) * sin(Pv12/2.0) * sin(Pv12/2.0); 
  Pv17 = sqrt( Pv16*Pv16 + pa * pa ); 

  if ( Pv13 - Pv17 == 0 ) 
    return 0.0; 
 
  if( Pv13 - Pv17 >= 0 ) 
    Pv18 = atan(sqrt(1.0-Pv13*Pv13)/Pv13); 
  else 
    Pv18 = atan(Pv17/sqrt(1.0-Pv17*Pv17)); 
  Pv20 = sin(Pv18); 
  Pv19 = cos(Pv18); 
  Pv21 = 1 - (Pv11*sin(Pv12)/Pv20)*(Pv11*sin(Pv12)/Pv20); 
  Pv22 = (L+L*L)*Pv18-Pv10*( L*L*Pv20/2.0 + (L*Pv18)*(L*Pv18) / Pv20); 
  Pv23 = Pv21*L*L*(Pv20*Pv19/4.0 - 5.0*(Pv18/4.0)+Pv18*Pv18*Pv19/Pv20); 
  Pv24 = (Pv22+Pv23)*(Pv11*sin(Pv12)/Pv20)+Pv12; 
  Pv25 = ( Pv15 + 2.0*cos(Pv9)*sin(Pv8)*sin(Pv24/2.0)*sin(Pv24/2.0) ) / 
         (cos(Pv9)*sin(Pv24) ); 
  Pv26 = (Pv15 - 2.0*cos(Pv8)*sin(Pv9)*sin(Pv24/2)*sin(Pv24/2) ) / 
         (cos(Pv8)*sin(Pv24)); 
 
  if( Pv25 == 0.0 ) 
    return 0.0; 

  if( Pv25 > 0.0 ) 
    Pv27 = PI - atan(-1.0/Pv25); 
  else 
    Pv27 = atan(1.0/Pv25); 
  if( Pv26 == 0.0 ) 
    return 0.0; 

  if( Pv26 > 0.0 ) 
    Pv28 = 2.0*PI - atan(-1.0/Pv26); 
  else 
    Pv28 = PI + atan(1.0/Pv26); 

  if( Pv5 < 0.0 ) 
    return 0.0; 

  if( Pv5 > 0.0 ) 
    return ( PI - Pv27 ); 

  return ( PI - Pv28 ); 
} 

